"""
Copyright (c) 2008-2021 synodriver <synodriver@gmail.com>
"""
from pybase16384.backends.cython._core_cy import (
    _decode,
    _decode_into,
    _encode,
    _encode_into,
    decode_file,
    decode_len,
    encode_file,
    encode_len,
    is_64bits,
)
