from typing import BinaryIO

def decode_len(dlen: int, offset: int) -> int: ...
def _decode(data: bytes) -> bytes: ...
def _decode_into(data: memoryview, dest: memoryview) -> int: ...
def decode_file(input: BinaryIO, output: BinaryIO, buf_rate: int = 10) -> None: ...
def encode_len(dlen: int) -> int: ...
def _encode(data: bytes) -> bytes: ...
def _encode_into(data: memoryview, dest: memoryview) -> int: ...
def encode_file(
    input: BinaryIO, output: BinaryIO, boolwrite_head: bool = False, buf_rate: int = 10
) -> None: ...
